import requests
from mcp.shared.exceptions import McpError
from mcp.types import (
    ErrorData,
    INTERNAL_ERROR,
    TextContent,
)
from jsonschema import validate


class HttpClient:
    api_url = "https://api.usaspending.gov"

    def __init__(self, endpoint, payload, response_schema):
        self.endpoint = endpoint
        self.payload = payload
        self.response_schema = response_schema

    def validate_response(self, response):
        if self.response_schema is None:
            return

        try:
            payload = response.json()
            validate(instance=payload, schema=self.response_schema)
        except Exception as e:
            print(f"Warning the response did not contain the expected information {e}")
            pass

    def handle_response(self, response) -> list[TextContent]:
        if response.status_code >= 200 and response.status_code < 300:
            print(response.text)
            self.validate_response(response)
            return [
                TextContent(
                    type="text",
                    text=response.text,
                )
            ]
        else:
            print(
                f"Non 2xx status code received {response.status_code} with response payload {response.text}"
            )
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,
                    message=f"Received unacceptable status code {response.status_code} with response payload {response.text}",
                )
            )


class PostClient(HttpClient):
    def __init__(self, endpoint, payload={}, response_schema=None):
        self.endpoint = endpoint
        self.payload = payload
        self.response_schema = response_schema

    def send(self):
        try:
            print(
                f"Sending POST request to {self.endpoint} with request payload {self.payload}"
            )
            response = requests.post(
                f"{self.api_url}{self.endpoint}", json=self.payload
            )
            return self.handle_response(response)
        except Exception as e:
            print(e)
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,
                    message=f"The following error occurred in the MCP server {e}",
                )
            )


class GetClient(HttpClient):
    def __init__(self, endpoint, params={}, response_schema=None):
        self.endpoint = endpoint
        self.params = params
        self.response_schema = response_schema

    def send(self):
        try:
            print(f"Sending GET request to {self.endpoint} with params {self.params}")
            response = requests.get(
                f"{self.api_url}{self.endpoint}", params=self.params
            )
            return self.handle_response(response)
        except Exception as e:
            print(e)
            raise McpError(
                ErrorData(
                    code=INTERNAL_ERROR,
                    message=f"The following error occurred in the MCP server {e}",
                )
            )



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import (
    ErrorData,
    INVALID_PARAMS,
    Tool,
)
from utils.http import GetClient

tool_awards_count_federal_accounts = Tool(
    name="awards_count_federal_accounts",
    description="Returns the number of federal accounts associated with the given award",
    inputSchema={
        "type": "object",
        "required": ["award_id"],
        "properties": {
            "award_id": {"type": "string", "description": "The name of the award"},
        },
    },
)

response_schema = {
    "type": "object",
    "properties": {
        "federal_accounts": {
            "type": "number",
        }
    },
}


def call_tool_awards_count_federal_accounts(arguments: dict[str, Any]):
    award_id = arguments.get("award_id")
    if not award_id:
        raise McpError(ErrorData(code=INVALID_PARAMS, message="award_id is required"))

    endpoint = f"/api/v2/awards/count/federal_account/{award_id}/"
    get_client = GetClient(endpoint=endpoint, response_schema=response_schema)
    return get_client.send()



from typing import Any
from mcp.types import Tool
from utils.http import PostClient

tool_recipient = Tool(
    name="recipient",
    description="This can be used to visualize the government spending that pertains to a specific recipient. This returns a list of recipients, their level, DUNS, UEI, and amount.",
    inputSchema={
        "type": "object",
        "required": [],
        "properties": {
            "order": {"type": "string", "enum": ["asc", "desc"]},
            "sort": {"type": "string", "enum": ["name", "duns", "amount"]},
            "limit": {"type": "number"},
            "page": {"type": "number"},
            "keyword": {
                "type": "string",
                "description": "They keyword results are filtered by. Searches on name, UEI, or DUNS",
            },
            "award_type": {
                "type": "string",
                "enum": [
                    "all",
                    "contracts",
                    "grants",
                    "loans",
                    "direct_payments",
                    "other_financial_assistance",
                ],
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": [],
    "properties": {
        "page_metadata": {
            "type": "object",
            "required": ["page", "limit", "total"],
            "properties": {
                "page": {"type": "number"},
                "limit": {"type": "number"},
                "total": {"type": "number"},
            },
        },
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": ["name", "duns", "uei", "id", "recipient_level"],
                "properties": {
                    "name": {
                        "type": ["string", "null"],
                        "description": "Name of the recipient. null when the name is not provided",
                    },
                    "duns": {
                        "type": ["string", "null"],
                        "description": "Recipient's DUNS (Data Universal Numbering System) number. null when no DUNS is provided",
                    },
                    "uei": {
                        "type": ["string", "null"],
                        "description": "Recipient's UEI (Unique Entity Identifier). null when no UEI is provided",
                    },
                    "amount": {
                        "type": "number",
                        "description": "The aggregate monetary value of all transactions associated with this recipient for the trailing 12 months.",
                    },
                    "recipient_level": {
                        "type": "string",
                        "enum": ["R", "P", "C"],
                        "description": "A letter representing the recipient level. R for neither parent nor child. P for Parent recipient, or C for child recipient",
                    },
                },
            },
        },
    },
}


def call_tool_recipient(arguments: dict[str, Any]):
    endpoint = "/api/v2/recipient/"
    order = arguments.get("order")
    sort = arguments.get("sort")
    limit = arguments.get("limit")
    page = arguments.get("page")
    keyword = arguments.get("keyword")
    award_type = arguments.get("award_type")

    payload = {}
    if order is not None:
        payload["order"] = order
    if sort is not None:
        payload["sort"] = sort
    if limit is not None:
        payload["limit"] = limit
    if page is not None:
        payload["page"] = page
    if keyword is not None:
        payload["keyword"] = keyword
    if award_type is not None:
        payload["award_type"] = award_type

    post_client = PostClient(endpoint, payload, response_schema)
    return post_client.send()



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import PostClient

tool_transactions = Tool(
    name="transactions",
    description="This returns a list of transactions, their amount, type, action date, action type, modification number, and description",
    inputSchema={
        "type": "object",
        "required": ["award_id"],
        "properties": {
            "award_id": {
                "type": "string",
                "description": "Either a generated natural id or a database surrogate award id. Generated award identifiers are preferred as they are effectively permanent. Surrogate award ids are retained for backward compatibility but are deprecated",
            },
            "limit": {"type": "number", "min": 1, "max": 5000},
            "page": {"type": "number", "default": 1},
            "sort": {
                "type": "string",
                "enum": [
                    "modification_number",
                    "action_date",
                    "federal_action_obligation",
                    "face_value_loan_guarantee",
                    "original_loan_subsidy_cost",
                    "action_type_description",
                    "description",
                ],
                "default": "action_date",
            },
            "order": {"type": "string", "enum": ["asc", "desc"], "default": "desc"},
        },
    },
)

response_schema = {
    "type": "object",
    "required": [],
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "action_date",
                    "action_type",
                    "action_type_description",
                    "description",
                    "face_value_loan_guarantee",
                    "federal_action_obligation",
                    "id",
                    "modification_number",
                    "original_loan_subsidy_cost",
                    "type",
                    "type_description",
                ],
                "properties": {
                    "action_date": {
                        "type": "string",
                        "description": "Action date in the format YYYY-MM-DD",
                    },
                    "action_type": {"type": "string"},
                    "action_type_description": {"type": ["string", "null"]},
                    "description": {"type": ["string", "null"]},
                    "face_value_loan_guarantee": {
                        "type": ["number", "null"],
                        "description": "Face value of the loan. Null for results with award type codes that do not correspond to loans",
                    },
                    "federal_action_obligation": {
                        "type": ["number", "null"],
                        "description": "Monetary value of the transaction. Null for results with award type codes that correspond to loans.",
                    },
                    "id": {"type": "string"},
                    "modification_number": {"type": "number"},
                    "original_loan_subsidy_cost": {
                        "type": ["number", "null"],
                        "description": "Original subsidy cost of the loan. Null for results with award type codes that do not correspond to loans.",
                    },
                    "type": {"type": "string"},
                    "type_description": {"type": ["string", "null"]},
                    "cfda_number": {"type": ["string", "null"]},
                },
            },
        },
        "page_metadata": {
            "type": "object",
            "required": ["page", "next", "previous", "hasNext", "hasPrevious"],
            "properties": {
                "page": {"type": "number"},
                "next": {"type": ["number", "null"]},
                "previous": {"type": ["number", "null"]},
                "hasNext": {"type": "boolean"},
                "hasPrevious": {"type": "boolean"},
            },
        },
    },
}


def call_tool_transactions(arguments: dict[str, Any]):
    endpoint = "/api/v2/transactions/"
    award_id = arguments.get("award_id")
    limit = arguments.get("limit")
    page = arguments.get("page")
    sort = arguments.get("sort")
    order = arguments.get("order")

    if award_id is None:
        raise McpError(
            ErrorData(code=INVALID_PARAMS, message="award_id is a required argument")
        )

    payload = {"award_id": award_id}

    if limit is not None:
        payload["limit"] = limit

    if page is not None:
        payload["page"] = page

    if sort is not None:
        payload["sort"] = sort

    if order is not None:
        payload["order"] = order

    post_client = PostClient(endpoint, payload, response_schema)
    return post_client.send()



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import GetClient

tool_major_object_class = Tool(
    name="major_object_class",
    description="This data can be used to better understand the different ways that a specific agency spends money",
    inputSchema={
        "type": "object",
        "required": ["fiscal_year", "funding_agency_id"],
        "properties": {
            "fiscal_year": {
                "type": "number",
                "description": "The fiscal year that you are querying data for",
            },
            "funding_agency_id": {
                "type": "number",
                "description": "The unique USAspending.gov agency identifier. This ID is the agency_id value returned in the /api/v2/references/toptier_agencies endpoint",
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": ["results"],
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "major_object_class_code",
                    "major_object_class_name",
                    "obligated_amount",
                ],
                "properties": {
                    "major_object_class_code": {"type": "string"},
                    "major_object_class_name": {"type": "string"},
                    "obligated_amount": {"type": "string"},
                },
            },
        }
    },
}


def call_tool_major_object_class(arguments: dict[str, Any]):
    fiscal_year = arguments.get("fiscal_year")
    funding_agency_id = arguments.get("funding_agency_id")
    if fiscal_year is None or funding_agency_id is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="fiscal_year and funding_agency_id must both be provided",
            )
        )

    params = {
        "fiscal_year": fiscal_year,
        "funding_agency_id": funding_agency_id,
    }

    endpoint = "/api/v2/financial_spending/major_object_class/"
    get_client = GetClient(endpoint, params, response_schema)
    return get_client.send()



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import PostClient

tool_subawards = Tool(
    name="subawards",
    description="This returns a filtered set of subawards",
    inputSchema={
        "type": "object",
        "required": ["page", "sort", "order"],
        "properties": {
            "page": {"type": "number", "default": 1},
            "limit": {"type": "number", "default": 10},
            "sort": {
                "type": "string",
                "enum": [
                    "subaward_number",
                    "id",
                    "description",
                    "action_date",
                    "amount",
                    "recipient_name",
                ],
            },
            "order": {"type": "string", "enum": ["asc", "desc"], "default": "desc"},
            "award_id": {
                "type": "string",
                "description": "Either a generated natural award id or a database surrogate award id. Generated award identifiers are preferred as they are effectively permanent. Surrogate award ids are retained for backward compatibility but are deprecated.",
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": ["results", "page_metadata"],
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "id",
                    "subaward_number",
                    "description",
                    "action_date",
                    "amount",
                    "recipient_name",
                ],
                "properties": {
                    "id": {"type": "number"},
                    "subaward_number": {"type": "string"},
                    "description": {"type": "string"},
                    "action_date": {"type": "string"},
                    "amount": {"type": "number"},
                    "recipient_name": {"type": "string"},
                },
            },
        },
        "page_metadata": {
            "type": "object",
            "required": ["page", "next", "previous", "hasNext", "hasPrevious"],
            "properties": {
                "page": {"type": "number"},
                "next": {"type": ["number", "null"]},
                "previous": {"type": ["number", "null"]},
                "hasNext": {"type": "boolean"},
                "hasPrevious": {"type": "boolean"},
            },
        },
    },
}


def call_tool_subawards(arguments: dict[str, Any]):
    endpoint = "/api/v2/subawards/"
    page = arguments.get("page")
    limit = arguments.get("limit")
    sort = arguments.get("sort")
    order = arguments.get("order")
    award_id = arguments.get("award_id")

    if page is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="page is a required argument",
            )
        )

    if sort is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="sort is a required argument",
            )
        )

    if order is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="order is a required argument",
            )
        )

    payload = {
        "page": page,
        "sort": sort,
        "order": order,
    }

    if limit is not None:
        payload["limit"] = limit

    if award_id is not None:
        payload["award_id"] = award_id

    post_client = PostClient(endpoint, payload, response_schema)
    return post_client.send()



import contextlib
import logging
from collections.abc import AsyncIterator
from typing import Any

import mcp.types as types
from mcp.server.lowlevel import Server
from mcp.server.streamable_http_manager import StreamableHTTPSessionManager
from starlette.applications import Starlette
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Mount
from starlette.types import Receive, Scope, Send

# Tools
from tools.v2.budget_functions.list_budget_functions import (
    call_tool_list_budget_functions,
    tool_list_budget_functions,
)
from tools.v2.federal_accounts import (
    call_tool_federal_accounts,
    tool_federal_accounts,
)
from tools.v2.financial_spending.major_object_class import (
    call_tool_major_object_class,
    tool_major_object_class,
)
from tools.v2.recipient import (
    call_tool_recipient,
    tool_recipient,
)
from tools.v2.references.toptier_agencies import (
    call_tool_toptier_agencies,
    tool_toptier_agencies,
)
from tools.v2.references.total_budgetary_resources import (
    call_tool_total_budgetary_resources,
    tool_total_budgetary_resources,
)
from tools.v2.search.spending_by_award import (
    tool_spending_by_award,
    call_tool_spending_by_award,
)
from tools.v2.spending import (
    call_tool_spending,
    tool_spending,
)
from tools.v2.subawards import (
    call_tool_subawards,
    tool_subawards,
)

logger = logging.getLogger(__name__)


def main() -> int:
    # Configure logging
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    app = Server("mcp-streamable-http-stateless-demo")

    @app.call_tool()
    async def call_tool(
        name: str, arguments: dict[str, Any]
    ) -> list[types.ContentBlock]:
        if name == "federal_accounts":
            return call_tool_federal_accounts(arguments)

        if name == "list_budget_functions":
            return call_tool_list_budget_functions()

        if name == "major_object_class":
            return call_tool_major_object_class(arguments)

        if name == "recipient":
            return call_tool_recipient(arguments)

        if name == "spending":
            return call_tool_spending(arguments)

        if name == "spending_by_award":
            return call_tool_spending_by_award(arguments)

        if name == "subawards":
            return call_tool_subawards(arguments)

        if name == "total_budgetary_resources":
            return call_tool_total_budgetary_resources(arguments)

        if name == "toptier_agencies":
            return call_tool_toptier_agencies(arguments)

        raise ValueError(f"Unknown tool: {name}")

    @app.list_tools()
    async def list_tools() -> list[types.Tool]:
        return [
            tool_federal_accounts,
            tool_list_budget_functions,
            tool_major_object_class,
            tool_recipient,
            tool_spending,
            tool_spending_by_award,
            tool_subawards,
            tool_total_budgetary_resources,
            tool_toptier_agencies,
        ]

    # Create the session manager with true stateless mode
    session_manager = StreamableHTTPSessionManager(
        app=app,
        event_store=None,
        stateless=True,
    )

    async def handle_streamable_http(
        scope: Scope, receive: Receive, send: Send
    ) -> None:
        await session_manager.handle_request(scope, receive, send)

    @contextlib.asynccontextmanager
    async def lifespan(app: Starlette) -> AsyncIterator[None]:
        """Context manager for session manager."""
        async with session_manager.run():
            logger.info("Application started with StreamableHTTP session manager!")
            try:
                yield
            finally:
                logger.info("Application shutting down...")

    # Create an ASGI application using the transport
    starlette_app = Starlette(
        debug=True,
        routes=[
            Mount("/mcp", app=handle_streamable_http),
        ],
        lifespan=lifespan,
    )

    # Wrap ASGI application with CORS middleware to expose Mcp-Session-Id header
    # for browser-based clients (ensures 500 errors get proper CORS headers)
    starlette_app = CORSMiddleware(
        starlette_app,
        allow_origins=["*"],  # Allow all origins - adjust as needed for production
        allow_methods=["GET", "POST", "DELETE"],  # MCP streamable HTTP methods
        expose_headers=["Mcp-Session-Id"],
    )

    import uvicorn

    uvicorn.run(starlette_app, host="127.0.0.1", port=8000)

    return 0


main()



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import GetClient

tool_federal_obligations = Tool(
    name="federal_obligations",
    description="This data can be used to better understand the different ways that a specific agency spends money. This returns the amount that the specific agency has obligated to various federal accounts in a given fiscal year.",
    inputSchema={
        "type": "object",
        "required": ["fiscal_year", "funding_agency_id"],
        "properties": {
            "fiscal_year": {
                "type": "number",
                "description": "The fiscal year that you are querying data for",
            },
            "funding_agency_id": {
                "type": "number",
                "description": "The unique USAspending.gov agency identifier. This ID is the agency_id value returned in the toptier_agencies tool",
            },
            "limit": {
                "type": "number",
                "description": "The maximum number of results to return in the response.",
            },
            "page": {
                "type": "number",
                "description": "The response page to return (the record offset is (page - 1) * limit)",
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": ["results", "page_metadata"],
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "account_title",
                    "account_number",
                    "id",
                    "obligated_amount",
                ],
                "properties": {
                    "account_title": {"type": "string"},
                    "account_number": {"type": "string"},
                    "id": {
                        "type": "string",
                        "description": "The USAspending.gov unique identifier for the federal account. You will need to use this ID when making API requests for details about specific federal accounts",
                    },
                    "obligated_amount": {"type": "string"},
                },
            },
        },
        "page_metadata": {
            "type": "object",
            "required": [
                "count",
                "page",
                "has_next_page",
                "has_previous_page",
                "next",
                "current",
                "previous",
            ],
            "properties": {
                "count": {"type": "number"},
                "page": {"type": "number"},
                "has_next_page": {"type": "boolean"},
                "has_previous_page": {"type": "boolean"},
                "next": {"type": ["string", "null"]},
                "current": {"type": "string"},
                "previous": {"type": ["string", "null"]},
            },
        },
    },
}


def call_tool_federal_obligations(arguments: dict[str, Any]):
    endpoint = "/api/v2/federal_obligations/"
    fiscal_year = arguments.get("fiscal_year")
    funding_agency_id = arguments.get("funding_agency_id")
    limit = arguments.get("limit")
    page = arguments.get("page")

    if fiscal_year is None:
        raise McpError(
            ErrorData(code=INVALID_PARAMS, message="fiscal_year must be provided")
        )

    if funding_agency_id is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="funding_agency_id must be provided",
            )
        )

    params = {
        "fiscal_year": fiscal_year,
        "funding_agency_id": funding_agency_id,
    }

    if limit is not None:
        params["limit"] = limit
    if page is not None:
        params["page"] = page

    get_client = GetClient(endpoint, params, response_schema)
    return get_client.send()



from typing import Any
from mcp.types import Tool
from utils.http import GetClient

tool_toptier_agencies = Tool(
    name="toptier_agencies",
    description="This data can be used to better understand the different ways that a specific agency spends money",
    inputSchema={
        "type": "object",
        "required": [],
        "properties": {
            "sort": {
                "type": "string",
                "description": "A data field used to sort the response array",
                "enum": [
                    "agency_id",
                    "agency_name",
                    "active_fy",
                    "active_fq",
                    "outlay_amount",
                    "obligated_amount",
                    "budget_authority_amount",
                    "current_total_budget_authority_amount",
                    "percentage_of_total_budget_authority",
                ],
            },
            "order": {
                "type": "string",
                "description": "The direction that the sort field will be sorted in",
                "enum": ["asc", "desc"],
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": ["results"],
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "abbreviation",
                    "active_fq",
                    "active_fy",
                    "agency_id",
                    "agency_name",
                    "budget_authority_amount",
                    "congressional_justification_url",
                    "current_total_budget_authority_amount",
                    "obligated_amount",
                    "outlay_amount",
                    "percentage_of_total_budget_authority",
                    "toptier_code",
                    "agency_slug",
                ],
                "properties": {
                    "abbreviation": {"type": "string"},
                    "active_fq": {"type": "string"},
                    "active_fy": {"type": "string"},
                    "agency_id": {
                        "type": "number",
                        "description": "The unique identifier for the agency. This is used in other endpoints when requesting detailed information about this specific agency",
                    },
                    "agency_name": {"type": "string"},
                    "budget_authority_amount": {"type": "number"},
                    "congressional_justification_url": {"type": ["string", "null"]},
                    "obligated_amount": {"type": "number"},
                    "outlay_amount": {"type": "number"},
                    "percentage_of_total_budget_authority": {
                        "type": "number",
                        "description": "This is the percentage of the agency's budget authority compared to the total budget authority",
                    },
                    "toptier_code": {"type": "string"},
                    "agency_slug": {
                        "type": "string",
                        "description": "The name of the agency in lowercase with dashed to be used for profile link construction",
                    },
                },
            },
        }
    },
}


def call_tool_toptier_agencies(arguments: dict[str, Any]):
    sort = arguments.get("sort")
    order = arguments.get("desc")
    params = {}
    if sort is not None:
        params["sort"] = sort
    if order is not None:
        params["order"] = order

    endpoint = "/api/v2/references/toptier_agencies/"
    get_client = GetClient(endpoint, params, response_schema)
    return get_client.send()



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import PostClient

standard_location_object = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["country"],
        "properties": {
            "country": {
                "type": "string",
                "description": "A 3 character code indication the country to search within. If the country code is not USA all further parameters can be ignored. A special country code FOREIGN represents all non-US countries.",
            },
            "state": {
                "type": "string",
                "description": "a 2 character string abbreviation for the state or territory",
                "minLength": 2,
                "maxLength": 2,
            },
            "county": {
                "type": "string",
                "description": "a 3 digit FIPS code indicating the county. If county is not provided, a state must be provided as well. If county is provided a district_original value must never be provided. If county is provided, a district_current value must never be provided.",
                "minLength": 3,
                "maxLength": 3,
            },
            "city": {
                "type": "string",
                "description": "If no state is provided, this will return results for all cities in any state with the provided name.",
            },
            "district_original": {
                "type": "string",
                "description": "A 2 character code indicating the congressional district. When provided a state must always be provided as well. When provided a county must never be provided. When provided, a country must always be USA. When provided a district_current must never be provided.",
                "minLength": 2,
                "maxLength": 2,
            },
            "district_current": {
                "type": "string",
                "description": "A 2 character code indicating the current congressional district. When provided a state must always be provided as well. When provided a county must never be provided. When provided a country must always be USA. When provided, a district_original value must never be provided.",
            },
            "zip": {
                "type": "string",
                "description": "A 5 digit string indicating the postal area to search within.",
            },
        },
    },
}

time_period_object = {
    "type": "array",
    "items": {
        "anyOf": [
            {
                "type": "object",
                "title": "SubawardSearchTimePeriodObject",
                "required": ["start_date", "end_date"],
                "description": "Use this if spending_level is subawards or subawards is true",
                "properties": {
                    "start_date": {
                        "type": "string",
                        "description": "Search based on one or more fiscal year selections OR date range. Dates should be in the following format: YYYY-MM-DD",
                    },
                    "end_date": {
                        "type": "string",
                        "description": "Search based on one or more fiscal year selections OR date range. Dates should be in the following format: YYYY-MM-DD",
                    },
                    "date_type": {
                        "type": "string",
                        "enum": ["action_date", "last_modified_date"],
                        "default": "action_date",
                    },
                },
            },
            {
                "type": "object",
                "title": "TransactionSearchTimePeriodObject",
                "required": ["start_date", "end_date"],
                "properties": {
                    "start_date": {
                        "type": "string",
                        "description": "Search based on one or more fiscal year selections OR date range. Dates should be in the following format: YYYY-MM-DD",
                    },
                    "end_date": {
                        "type": "string",
                        "description": "Search based on one or more fiscal year selections OR date range. Dates should be in the following format: YYYY-MM-DD",
                    },
                    "date_type": {
                        "type": "string",
                        "enum": [
                            "action_date",
                            "date_signed",
                            "last_modified_date",
                            "new_awards_only",
                        ],
                    },
                },
            },
        ]
    },
}

filter_object_award_types = {
    "type": "array",
    "default": ["A", "B", "C", "D"],
    "items": {
        "type": "string",
        "enum": [
            "02",
            "03",
            "04",
            "05",
            "06",
            "07",
            "08",
            "09",
            "10",
            "11",
            "A",
            "B",
            "C",
            "D",
            "IDV_A",
            "IDV_B",
            "IDV_B_A",
            "IDV_B_B",
            "IDV_B_C",
            "IDV_C",
            "IDV_D",
            "IDV_E",
        ],
    },
}

spending_by_award_fields_enum = [
    "Award ID",
    "Recipient Name",
    "Recipient DUNS Number",
    "recipient_id",
    "Awarding Agency",
    "Awarding Agency Code",
    "Awarding Sub Agency",
    "Awarding Sub Agency Code",
    "Funding Agency",
    "Funding Agency Code",
    "Funding Sub Agency",
    "Funding Sub Agency Code",
    "Place of Performance City Code",
    "Place of Performance State Code",
    "Place of Performance Country Code",
    "Place of Performance Zip5",
    "Description",
    "Last Modified Date",
    "Base Obligation Date",
    "prime_award_recipient_id",
    "generated_internal_id",
    "def_codes",
    "COVID-19 Obligations",
    "COVID-19 Outlays",
    "Infrastructure Obligations",
    "Infrastructure Outlays",
    "Recipient UEI",
    "Recipient Location",
    "Primary Place of Performance",
]


spending_by_award_fields = {
    "type": "array",
    "default": spending_by_award_fields_enum,
    "items": {
        "type": "string",
    },
}

# Used in the request payload to sort results
spending_by_award_response_properties = [
    "internal_id",
    "Award Amount",
    "Total Outlays",
    "Award ID",
    "Award Type",
    "Awarding Agency Code",
    "Awarding Agency",
    "awarding_agency_id",
    "Awarding Sub Agency CodeAwarding Sub Agency",
    "Base Obligation DateCFDA NumberContract Award Type",
    "Description",
    "End Date",
    "Funding Agency Code",
    "Funding Agency",
    "Funding Sub Agency Code",
    "Funding Sub Agency",
    "generated_internal_id",
    "Issued Date",
    "Last Date to Order",
    "Last Modified Date",
    "Loan Value",
    "Period of Performance Current End Date",
    "Period of Performance Start Date",
    "Place of Performance City Code",
    "Place of Performance Country Code",
    "Place of Performance State Code",
    "Place of Performance Zip5",
    "COVID-19 Outlays",
    "COVID-19 Obligations",
    "Infrastructure Outlays",
    "Infrastructure Obligations",
    "def_codes",
    "Prime Award ID",
    "Prime Recipient Name",
    "prime_award_recipient_id",
    "prime_award_internal_id",
    "prime_award_generated_internal_id",
    "Recipient DUNS Number",
    "Recipient Name",
    "recipient_id",
    "SAI Number",
    "Start Date",
    "Sub-Award Amount",
    "Sub-Award Date",
    "Sub-Award ID",
    "Sub-Award Type",
    "Sub-Awardee Name",
    "Subsidy Cost",
    "agency_slug",
]

"""
We are going to take a different approach for this tool versus spending_by_geography
I am going to reduce the input schema since I am worried a more complex input schema may overwhelm the all knowing AI
Weird, initially I didn't event supply recipient_search_text but it still tried to use it in a query so I added that one since it tried using it as a string instead of an array
"""
tool_spending_by_award = Tool(
    name="spending_by_award",
    description="This allows for complex filtering for specific subsets of spending data. This accepts filters and fields, and returns the fields of the filtered awards.",
    inputSchema={
        "type": "object",
        "required": ["filters", "fields"],
        "properties": {
            "filters": {
                "type": "object",
                "required": ["award_type_codes"],
                "properties": {
                    "time_period": time_period_object,
                    "place_of_performance_locations": standard_location_object,
                    "award_type_codes": filter_object_award_types,
                    "recipient_search_text": {
                        "type": "array",
                        "description": "Text searched across a recipients name, UEI, and DUNS",
                        "minItems": 1,  # Will return 422 if this is below min 1 items
                        "items": {
                            "type": "string",
                        },
                    },
                },
            },
            "fields": spending_by_award_fields,
            "limit": {
                "type": "number",
                "description": "How many results are returned.",
                "default": 10,
            },
            "order": {
                "type": "string",
                "enum": ["asc", "desc"],
                "default": "desc",
                "description": "Indicates what direction results should be sorted by.",
            },
            "sort": {
                "type": "string",
                "description": "What value the results should be sorted by.",
                "enum": spending_by_award_response_properties,
            },
            "page": {"type": "string"},
            "subawards": {
                "type": "boolean",
                "description": "True when you want to group by Subawards instead of Awards",
                "default": False,
            },
            "spending_level": {
                "type": "string",
                "description": "Group the spending by level. This also determines what data source is used for the totals",
                "enum": ["awards", "subawards"],
                "default": "awards",
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": ["spending_level", "limit", "results"],
    "properties": {
        "spending_level": {
            "type": "string",
            "enum": ["awards", "subawards"],
        },
        "limit": {"type": "number"},
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": ["internal_id"],
                "properties": {
                    "internal_id": {"type": "number"},
                    "Award Amount": {"type": "number"},
                    "Total Outlays": {"type": "number"},
                    "Award ID": {"type": "string"},
                    "Award Type": {"type": ["string", "null"]},
                    "Awarding Agency Code": {"type": ["string", "null"]},
                    "Awarding Agency": {"type": ["string", "null"]},
                    "awarding_agency_id": {"type": ["number", "null"]},
                    "Awarding Sub Agency Code": {"type": ["string", "null"]},
                    "Awarding Sub Agency": {"type": ["string", "null"]},
                    "Base Obligation Date": {"type": "string"},
                    "CFDA Number": {"type": ["string", "null"]},
                    "Contract Award Type": {"type": "string"},
                    "Description": {"type": ["string", "null"]},
                    "End Date": {"type": "string"},
                    "Funding Agency Code": {"type": ["string", "null"]},
                    "Funding Agency": {"type": ["string", "null"]},
                    "Funding Sub Agency Code": {"type": ["string", "null"]},
                    "Funding Sub Agency": {"type": ["string", "null"]},
                    "generated_internal_id": {"type": "string"},
                    "Issued Date": {"type": ["string", "null"]},
                    "Last Date to Order": {"type": ["string", "null"]},
                    "Last Modified Date": {"type": "string"},
                    "Loan Value": {"type": ["number"]},
                    "Period of Performance Current End Date": {
                        "type": ["string", "null"]
                    },
                    "Period of Performance Start Date": {"type": "string"},
                    "Place of Performance City Code": {"type": "number"},
                    "Place of Performance Country Code": {"type": ["string", "null"]},
                    "Place of Performance State Code": {"type": ["number", "null"]},
                    "Place of Performance Zip5": {"type": "number"},
                    "COVID-19 Outlays": {"type": ["number"]},
                    "COVID-19 Obligations": {"type": ["number"]},
                    "Infrastructure Outlays": {"type": ["number"]},
                    "Infrastructure Obligations": {"type": ["number"]},
                    "def_codes": {"type": "array", "items": {"type": "string"}},
                    "Prime Award ID": {"type": ["string", "null"]},
                    "Prime Recipient Name": {"type": ["string", "null"]},
                    "prime_award_recipient_id": {"type": ["string", "null"]},
                    "prime_award_internal_id": {"type": ["string", "null"]},
                    "prime_award_generated_internal_id": {"type": ["string"]},
                    "Recipient DUNS Number": {"type": ["string", "null"]},
                    "Recipient Name": {"type": ["string", "null"]},
                    "recipient_id": {"type": ["string", "null"]},
                    "SAI Number": {"type": ["string", "null"]},
                    "Start Date": {"type": "string"},
                    "Sub-Award Amount": {"type": ["string"]},
                    "Sub-Award Date": {"type": ["string"]},
                    "Sub-Award ID": {"type": ["string"]},
                    "Sub-Award Type": {"type": ["string"]},
                    "Sub-Awardee Name": {"type": ["string"]},
                    "Subsidy Cost": {"type": ["number"]},
                    "agency_slug": {"type": ["string", "null"]},
                },
            },
        },
        "page_metadata": {
            "type": "object",
            "required": ["page", "hasNext"],
            "properties": {
                "page": {"type": "number"},
                "hasNext": {"type": "boolean"},
            },
        },
        "messages": {"type": "array", "items": {"type": "string"}},
    },
}


def call_tool_spending_by_award(arguments: dict[str, Any]):
    endpoint = "/api/v2/search/spending_by_award/"
    filters = arguments.get("filters")
    fields = arguments.get("fields")
    limit = arguments.get("limit")
    order = arguments.get("order")
    page = arguments.get("page")
    sort = arguments.get("sort")
    subawards = arguments.get("subawards")
    spending_level = arguments.get("spending_level")

    if not bool(filters) or fields is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="filters and fields are required arguments",
            )
        )

    payload = {
        "filters": filters,
        "fields": fields,
    }

    if limit is not None:
        payload["limit"] = limit
    if order is not None:
        payload["order"] = order
    if page is not None:
        payload["page"] = page
    if sort is not None:
        payload["sort"] = sort
    if subawards is not None:
        payload["subawards"] = subawards
    if spending_level is not None:
        payload["spending_level"] = spending_level

    post_client = PostClient(endpoint, payload, response_schema)
    return post_client.send()



from mcp.types import Tool
from utils.http import GetClient

tool_list_budget_functions = Tool(
    name="list_budget_functions",
    description="This retrieves a list of all Budget Functions ordered by their title",
    inputSchema={"type": "object", "properties": {}},
)

response_schema = {
    "type": "object",
    "required": ["results"],
    "properties": {
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": ["budget_function_code", "budget_function_title"],
                "properties": {
                    "budget_function_code": {"type": "string"},
                    "budget_function_title": {"type": "string"},
                },
            },
        }
    },
}


def call_tool_list_budget_functions():
    endpoint = "/api/v2/budget_functions/list_budget_functions/"
    get_client = GetClient(endpoint=endpoint, response_schema=response_schema)
    return get_client.send()



from typing import Any
from mcp.types import Tool
from utils.http import PostClient

tool_federal_accounts = Tool(
    name="federal_accounts",
    description="This returns a list of federal accounts, their number, name, managing agency, and budgetary resources",
    inputSchema={
        "type": "object",
        "required": [],
        "properties": {
            "filters": {
                "type": "object",
                "required": [],
                "properties": {
                    "fy": {"type": "string"},
                    "agency_identifier": {"type": "string"},
                },
            },
            "sort": {
                "type": "object",
                "required": [],
                "properties": {
                    "direction": {
                        "type": "string",
                        "enum": ["asc", "desc"],
                    }
                },
            },
            "limit": {
                "type": "number",
                "description": "The number of results to include per page. The default is 50",
            },
            "keyword": {
                "type": "string",
                "description": "They keyword that you want to search on. Can be used to search by name, number, managing agency, and budgetary resources",
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": [
        "count",
        "limit",
        "hasNext",
        "page",
        "hasPrevious",
        "next",
        "fy",
        "results",
    ],
    "properties": {
        "previous": {"type": ["number", "null"]},
        "count": {"type": "number"},
        "limit": {"type": "number"},
        "hasNext": {"type": "boolean"},
        "page": {"type": "number"},
        "hasPrevious": {"type": "boolean"},
        "next": {"type": ["number", "null"]},
        "fy": {"type": "string"},
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "account_name",
                    "account_number",
                    "account_id",
                    "managing_agency_acronym",
                    "agency_identifier",
                    "budgetary_resources",
                    "managing_agency",
                ],
                "properties": {
                    "account_name": {
                        "type": ["string", "null"],
                        "description": "Name of the federal account. null when the name is not provided",
                    },
                    "account_number": {
                        "type": ["string", "null"],
                        "description": "The number for the federal account. null when no account number is provided",
                    },
                    "account_id": {
                        "type": "number",
                        "description": "A unique identifier for the federal account",
                    },
                    "managing_agency_acronym": {"type": "string"},
                    "managing_agency": {"type": "string"},
                },
            },
        },
    },
}


def call_tool_federal_accounts(arguments: dict[str, Any]):
    endpoint = "/api/v2/federal_accounts/"
    filters = arguments.get("filters")
    sort = arguments.get("sort")
    limit = arguments.get("limit")
    page = arguments.get("page")
    keyword = arguments.get("keyword")

    payload = {}
    if bool(filters):
        payload["filters"] = filters
    if bool(sort):
        payload["sort"] = sort
    if limit is not None:
        payload["limit"] = limit
    if page is not None:
        payload["page"] = page
    if keyword is not None:
        payload["keyword"] = keyword

    post_client = PostClient(endpoint, payload, response_schema)
    return post_client.send()



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import GetClient

tool_total_budgetary_resources = Tool(
    name="total_budgetary_resources",
    description="This is used to provide information on the federal budgetary resources of the government",
    inputSchema={
        "type": "object",
        "required": [],
        "properties": {
            "fiscal_year": {
                "type": "number",
                "description": "The fiscal year to retrieve, 2017 or later",
            },
            "fiscal_period": {
                "type": "number",
                "description": "The fiscal period",
            },
        },
    },
)

response_schema = {
    "type": "object",
    "properties": {
        "messages": {"type": "array", "items": {"type": "string"}},
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": [
                    "fiscal_year",
                    "fiscal_period",
                    "total_budgetary_resources",
                ],
                "properties": {
                    "fiscal_year": {
                        "type": "number",
                    },
                    "fiscal_period": {
                        "type": "number",
                    },
                    "total_budgetary_resources": {"type": "number"},
                },
            },
        },
    },
}


def call_tool_total_budgetary_resources(arguments: dict[str, Any]):
    fiscal_year = arguments.get("fiscal_year")
    fiscal_period = arguments.get("fiscal_period")
    if fiscal_period is not None and fiscal_year is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="If fiscal_period is provided then fiscal_year must be provided as well",
            )
        )

    params = {}
    if fiscal_year is not None:
        params["fiscal_year"] = fiscal_year
    if fiscal_period is not None:
        params["fiscal_period"] = fiscal_period

    endpoint = "/api/v2/references/total_budgetary_resources/"
    get_client = GetClient(endpoint, params, response_schema)
    return get_client.send()



from mcp.types import Tool
from utils.util import send

tool_state = Tool(
    name="state",
    description="This data contains the government spending that occurs in a specific state or territory",
    inputSchema={"type": "object", "properties": {}},
)

response_schema = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["name", "code", "fips", "amount", "type"],
        "properties": {
            "name": {"type": "string"},
            "code": {"type": "string"},
            "fips": {"type": "string"},
            "type": {"type": "string", "enum": ["state", "territory", "district"]},
        },
    },
}


def call_tool_state():
    endpoint = "/api/v2/recipient/state/"
    return send(endpoint=endpoint, response_schema=response_schema)



from typing import Any
from mcp.shared.exceptions import McpError
from mcp.types import ErrorData, INVALID_PARAMS, Tool
from utils.http import PostClient

tool_spending = Tool(
    name="spending",
    description="This data can be used to drill down into specific subsets of data by level of detail. This data represents all government spending in the specified time period, grouped by the data type of your choice.",
    inputSchema={
        "type": "object",
        "required": ["type", "filters"],
        "properties": {
            "type": {
                "type": "string",
                "enum": [
                    "federal_account",
                    "object_class",
                    "recipient",
                    "award",
                    "budget_function",
                    "budget_subfunction",
                    "agency",
                    "program_activity",
                ],
            },
            "filters": {
                "type": "object",
                "required": ["fy"],
                "properties": {
                    "fy": {"type": "string"},
                    "quarter": {"type": "string", "enum": ["1", "2", "3", "4"]},
                    "period": {
                        "type": "string",
                        "enum": [
                            "1",
                            "2",
                            "3",
                            "4",
                            "5",
                            "6",
                            "7",
                            "8",
                            "9",
                            "10",
                            "11",
                            "12",
                        ],
                    },
                    "agency": {"type": "number"},
                    "federal_account": {"type": "number"},
                    "object_class": {"type": "number"},
                    "budget_function": {"type": "number"},
                    "budget_subfunction": {"type": "number"},
                    "recipient": {"type": "number"},
                    "program_activity": {"type": "number"},
                },
            },
        },
    },
)

response_schema = {
    "type": "object",
    "required": ["total", "end_date", "results"],
    "properties": {
        "total": {"type": ["number", "null"]},
        "end_date": {"type": "string"},
        "results": {
            "type": "array",
            "items": {
                "type": "object",
                "required": ["code", "id", "type", "name", "amount"],
                "properties": {
                    "code": {"type": "string"},
                    "id": {"type": "string"},
                    "generated_unique_number_id": {"type": "string"},
                    "type": {"type": "string"},
                    "name": {"type": "string"},
                    "amount": {"type": "number"},
                    "account_number": {"type": "string"},
                    "link": {"type": "string"},
                },
            },
        },
    },
}


def call_tool_spending(arguments: dict[str, Any]):
    endpoint = "/api/v2/spending/"
    spending_type = arguments.get("type")
    filters = arguments.get("filters")
    if spending_type is None:
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="type is a required argument",
            )
        )

    if not bool(filters):
        raise McpError(
            ErrorData(
                code=INVALID_PARAMS,
                message="filters is a required argument",
            )
        )

    payload = {
        "type": spending_type,
        "filters": filters,
    }

    post_client = PostClient(endpoint, payload, response_schema)
    return post_client.send()
